@@ -33,6 +33,9 @@ except NameError:
         pass
 
 
+_VALID_FILE_MODES = {'r', 'w', 'a', 'rb', 'wb', 'ab'}
+
+
 def tokenize(*args, **kwargs):
     """ Deterministic token
 
@@ -112,10 +115,13 @@ class S3FileSystem(object):
         ``S3File.open``.
     version_aware : bool (False)
         Whether to support bucket versioning.  If enable this will require the
-        user to have the neccesary IAM permissions for dealing with versioned
+        user to have the necessary IAM permissions for dealing with versioned
         objects.
     config_kwargs : dict of parameters passed to ``botocore.client.Config``
     kwargs : other parameters for boto3 session
+    session : botocore Session object to be used for all connections.
+         This session will be used inplace of creating a new session inside S3FileSystem.
+
 
     Examples
     --------
@@ -137,9 +143,12 @@ class S3FileSystem(object):
                  use_ssl=True, client_kwargs=None, requester_pays=False,
                  default_block_size=None, default_fill_cache=True,
                  version_aware=False, config_kwargs=None,
-                 s3_additional_kwargs=None, **kwargs):
+                 s3_additional_kwargs=None, session=None, **kwargs):
         self.anon = anon
         self.session = None
+        self.passed_in_session = session
+        if self.passed_in_session:
+            self.session = self.passed_in_session
         self.key = key
         self.secret = secret
         self.token = token
@@ -147,10 +156,9 @@ class S3FileSystem(object):
 
         if client_kwargs is None:
             client_kwargs = {}
-        if default_block_size is not None:
-            self.default_block_size = default_block_size
         if config_kwargs is None:
             config_kwargs = {}
+        self.default_block_size = default_block_size or self.default_block_size
         self.default_fill_cache = default_fill_cache
         self.version_aware = version_aware
         self.client_kwargs = client_kwargs
@@ -203,7 +211,7 @@ class S3FileSystem(object):
         anon, key, secret, kwargs, ckwargs, token, ssl = (
               self.anon, self.key, self.secret, self.kwargs,
               self.client_kwargs, self.token, self.use_ssl)
-
+        
         # Include the current PID in the connection key so that different
         # SSL connections are made for each process.
         tok = tokenize(anon, key, secret, kwargs, ckwargs, token,
@@ -212,24 +220,30 @@ class S3FileSystem(object):
             self._conn.pop(tok, None)
         if tok not in self._conn:
             logger.debug("Open S3 connection.  Anonymous: %s", self.anon)
+
             if self.anon:
                 from botocore import UNSIGNED
                 conf = Config(connect_timeout=self.connect_timeout,
                               read_timeout=self.read_timeout,
                               signature_version=UNSIGNED, **self.config_kwargs)
-                self.session = boto3.Session(**self.kwargs)
+                if not self.passed_in_session:
+                    self.session = boto3.Session(**self.kwargs)
             else:
                 conf = Config(connect_timeout=self.connect_timeout,
                               read_timeout=self.read_timeout,
                               **self.config_kwargs)
-                self.session = boto3.Session(self.key, self.secret, self.token,
-                                             **self.kwargs)
+                if not self.passed_in_session:
+                    self.session = boto3.Session(self.key, self.secret, self.token,
+                                                 **self.kwargs)
+
             s3 = self.session.client('s3', config=conf, use_ssl=ssl,
                                      **self.client_kwargs)
             self._conn[tok] = (s3, self.session)
         else:
             s3, session = self._conn[tok]
-            self.session = session
+            if not self.passed_in_session:
+                self.session = session
+
         return s3
 
     def get_delegated_s3pars(self, exp=3600):
@@ -258,6 +272,8 @@ class S3FileSystem(object):
                 'token': cred['SessionToken'], 'anon': False}
 
     def __getstate__(self):
+        if self.passed_in_session:
+            raise NotImplementedError
         d = self.__dict__.copy()
         del d['s3']
         del d['session']
@@ -271,7 +287,7 @@ class S3FileSystem(object):
         self._kwargs_helper = ParamKwargsHelper(self.s3)
 
     def open(self, path, mode='rb', block_size=None, acl='', version_id=None,
-             fill_cache=None, **kwargs):
+             fill_cache=None, encoding=None, **kwargs):
         """ Open a file for reading or writing
 
         Parameters
@@ -279,7 +295,8 @@ class S3FileSystem(object):
         path: string
             Path of file on S3
         mode: string
-            One of 'rb' or 'wb'
+            One of 'r', 'w', 'a', 'rb', 'wb', or 'ab'. These have the same meaning
+            as they do for the built-in `open` function.
         block_size: int
             Size of data-node blocks if reading
         fill_cache: bool
@@ -292,7 +309,10 @@ class S3FileSystem(object):
         version_id : str
             Explicit version of the object to open.  This requires that the s3
             filesystem is version aware and bucket versioning is enabled on the
-            relavant bucket.
+            relevant bucket.
+        encoding : str
+            The encoding to use if opening the file in text mode. The platform's
+            default text encoding is used if not given.
         kwargs: dict-like
             Additional parameters used for s3 methods.  Typically used for
             ServerSideEncryption.
@@ -301,18 +321,21 @@ class S3FileSystem(object):
             block_size = self.default_block_size
         if fill_cache is None:
             fill_cache = self.default_fill_cache
-        if 'b' not in mode:
-            raise NotImplementedError("Text mode not supported, use mode='%s'"
-                                      " and manage bytes" % (mode[0] + 'b'))
+
         acl = acl or self.s3_additional_kwargs.get('ACL', '')
         kw = self.s3_additional_kwargs.copy()
         kw.update(kwargs)
         if not self.version_aware and version_id:
             raise ValueError("version_id cannot be specified if the filesystem "
                              "is not version aware")
-        return S3File(self, path, mode, block_size=block_size, acl=acl,
-                      version_id=version_id, fill_cache=fill_cache,
-                      s3_additional_kwargs=kw)
+
+        mode2 = mode if 'b' in mode else (mode.replace('t', '') + 'b')
+        fdesc = S3File(self, path, mode2, block_size=block_size, acl=acl,
+                       version_id=version_id, fill_cache=fill_cache,
+                       s3_additional_kwargs=kw)
+        if 'b' in mode:
+            return fdesc
+        return io.TextIOWrapper(fdesc, encoding=encoding)
 
     def _lsdir(self, path, refresh=False):
         if path.startswith('s3://'):
@@ -763,7 +786,15 @@ class S3FileSystem(object):
                     f2.write(data)
 
     def mkdir(self, path, acl="", **kwargs):
-        """ Make new bucket or empty key """
+        """ Make new bucket or empty key
+
+        Parameters
+        ----------
+        acl: str
+            ACL to set when creating
+        region_name : str
+            region in which the bucket should be created
+        """
         acl = acl or self.s3_additional_kwargs.get('ACL', '')
         self.touch(path, acl=acl, **kwargs)
 
@@ -952,7 +983,14 @@ class S3FileSystem(object):
             if acl and acl not in buck_acls:
                 raise ValueError('ACL not in %s', buck_acls)
             try:
-                self.s3.create_bucket(Bucket=bucket, ACL=acl)
+                params = {"Bucket": bucket, 'ACL': acl}
+                region_name = (kwargs.get("region_name", None) or
+                               self.client_kwargs.get("region_name", None))
+                if region_name:
+                    params['CreateBucketConfiguration'] = {
+                        'LocationConstraint': region_name
+                    }
+                self.s3.create_bucket(**params)
                 self.invalidate_cache('')
                 self.invalidate_cache(bucket)
             except (ClientError, ParamValidationError):
@@ -1015,6 +1053,9 @@ class S3File(object):
         boto3 connection
     path : string
         S3 bucket/key to access
+    mode : str
+        One of 'rb', 'wb', 'ab'. These have the same meaning
+        as they do for the built-in `open` function.
     block_size : int
         read-ahead size for finding delimiters
     fill_cache : bool
@@ -1065,7 +1106,7 @@ class S3File(object):
         self.acl = acl
         self.fill_cache = fill_cache
         self.s3_additional_kwargs = s3_additional_kwargs or {}
-        if mode in {'wb', 'ab'}:
+        if self.writable():
             if acl and acl not in key_acls:
                 raise ValueError('ACL not in %s', key_acls)
             self.buffer = io.BytesIO()
@@ -1074,7 +1115,7 @@ class S3File(object):
             if block_size < 5 * 2 ** 20:
                 raise ValueError('Block size must be >=5MB')
             self.forced = False
-            if mode == 'ab' and s3.exists(path):
+            if 'a' in mode and s3.exists(path):
                 self.size = s3.info(path)['Size']
                 if self.size < 5*2**20:
                     # existing file too small for multi-upload: download
@@ -1146,7 +1187,7 @@ class S3File(object):
         --------
         >>> mys3file.setxattr(attribute_1='value1', attribute_2='value2')  # doctest: +SKIP
         """
-        if self.mode != 'rb':
+        if self.writable():
             raise NotImplementedError('cannot update metadata while file '
                                       'is open for writing')
         return self.s3.setxattr(self.path, copy_kwargs=copy_kwargs, **kwargs)
@@ -1170,7 +1211,7 @@ class S3File(object):
         whence : {0, 1, 2}
             from start of file, current location or end of file, resp.
         """
-        if not self.mode == 'rb':
+        if not self.readable():
             raise ValueError('Seek only available in read mode')
         if whence == 0:
             nloc = loc
@@ -1261,7 +1302,7 @@ class S3File(object):
         length : int (-1)
             Number of bytes to read; if <0, all remaining bytes.
         """
-        if self.mode != 'rb':
+        if not self.readable():
             raise ValueError('File not in read mode')
         if length < 0:
             length = self.size
@@ -1290,7 +1331,7 @@ class S3File(object):
         data : bytes
             Set of bytes to be written.
         """
-        if self.mode not in {'wb', 'ab'}:
+        if not self.writable():
             raise ValueError('File not in write mode')
         if self.closed:
             raise ValueError('I/O operation on closed file.')
@@ -1317,7 +1358,7 @@ class S3File(object):
             blocks are allowed to be.
         retries: int
         """
-        if self.mode in {'wb', 'ab'} and not self.closed:
+        if self.writable() and not self.closed:
             if self.buffer.tell() < self.blocksize and not force:
                 # ignore if not enough data for a block and not closing
                 return
@@ -1371,7 +1412,7 @@ class S3File(object):
         if self.closed:
             return
         self.cache = None
-        if self.mode in {'wb', 'ab'}:
+        if self.writable():
             if self.parts:
                 self.flush(force=True)
                 part_info = {'Parts': self.parts}
@@ -1382,22 +1423,22 @@ class S3File(object):
                     UploadId=self.mpu['UploadId'],
                     MultipartUpload=part_info)
             else:
-                self.buffer.seek(0)
                 try:
                     self._call_s3(
                         self.s3.s3.put_object,
                         Bucket=self.bucket, Key=self.key,
-                        Body=self.buffer.read(), ACL=self.acl)
+                        Body=self.buffer.getvalue(), ACL=self.acl)
                 except (ClientError, ParamValidationError) as e:
                     raise IOError('Write failed: %s' % self.path, e)
             self.s3.invalidate_cache(self.path)
             self.parts = []
-            self.buffer = None
+            self.buffer.seek(0)
+            self.buffer.truncate(0)
         self.closed = True
 
     def readable(self):
         """Return whether the S3File was opened for reading"""
-        return self.mode == 'rb'
+        return 'r' in self.mode
 
     def seekable(self):
         """Return whether the S3File is seekable (only in read mode)"""
@@ -1405,7 +1446,7 @@ class S3File(object):
 
     def writable(self):
         """Return whether the S3File was opened for writing"""
-        return self.mode in {'wb', 'ab'}
+        return 'w' in self.mode or 'a' in self.mode
 
     def __del__(self):
         self.close()
@@ -1421,6 +1462,22 @@ class S3File(object):
     def __exit__(self, *args):
         self.close()
 
+    # Implementations of BufferedIOBase stub methods
+
+    def read1(self, length=-1):
+        return self.read(length)
+
+    def detach(self):
+        raise io.UnsupportedOperation()
+
+    def readinto(self, b):
+        data = self.read()
+        b[:len(data)] = data
+        return len(data)
+
+    def readinto1(self, b):
+        return self.readinto(b)
+
 
 def _fetch_range(client, bucket, key, version_id, start, end, max_attempts=10,
                  req_kw=None):
